import os
import logging
import logging.config
from typing import Literal
from pathlib import Path
from httpx import ConnectError

from pydantic import BaseSettings, Field, validator, HttpUrl, Extra

from ixmp4.core.exceptions import InvalidCredentials
from .credentials import Credentials
from .toml import TomlConfig
from .manager import ManagerConfig
from .auth import ManagerAuth, AnonymousAuth
from .user import local_user

logger = logging.getLogger(__name__)

here = Path(__file__).parent


class Settings(BaseSettings):
    mode: Literal["production"] | Literal["development"] | Literal[
        "debug"
    ] = "production"
    storage_directory: Path = Field("~/.local/share/ixmp4/", env="ixmp4_dir")
    secret_hs256: str = "default_secret_hs256"
    migration_db_uri: str = "sqlite:///./run/db.sqlite"
    manager_url: HttpUrl = Field("https://api.manager.ece.iiasa.ac.at/v1")
    managed: bool = True

    class Config:
        env_prefix = "ixmp4_"
        extra = Extra.allow

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.storage_directory.mkdir(parents=True, exist_ok=True)

        self.database_dir = self.storage_directory / "databases"
        self.database_dir.mkdir(exist_ok=True)

        self.log_dir = self.storage_directory / "log"
        self.log_dir.mkdir(exist_ok=True)

        self.configure_logging(self.mode)

        self.load_credentials()
        self.load_manager_config()
        self.load_toml_config()

    def load_credentials(self):
        credentials_config = self.storage_directory / "credentials.toml"
        credentials_config.touch()
        self.credentials = Credentials(credentials_config)

        self.default_credentials = None
        self.default_auth = None
        try:
            self.default_credentials = self.credentials.get("default")
        except KeyError:
            logger.warn("No default credentials provided.")

        if self.default_credentials is not None:
            username, password = self.default_credentials
            try:
                self.default_auth = ManagerAuth(username, password, self.manager_url)
                return
            except InvalidCredentials:
                logger.warn(
                    "Failure while requesting management service authentication: Invalid credentials."
                )
            except ConnectError:
                logger.warn(f"Unable to connect to {self.manager_url}.")

        self.default_auth = AnonymousAuth()

    def load_manager_config(self):
        self.manager = None
        if self.default_auth is not None:
            self.manager = ManagerConfig(
                self.manager_url, self.default_auth, remote=True
            )

    def load_toml_config(self):
        toml_user = self.default_auth.get_user()
        if not toml_user.is_authenticated:
            toml_user = local_user

        toml_config = self.storage_directory / "platforms.toml"
        toml_config.touch()
        self.toml = TomlConfig(toml_config, toml_user)

    @validator("storage_directory")
    def expand_user(cls, v):
        # translate ~/asdf into /home/user/asdf
        return Path.expanduser(v)

    def configure_logging(self, config: str):
        access_file = self.log_dir / "access.log"
        debug_file = self.log_dir / "debug.log"
        error_file = self.log_dir / "error.log"
        os.environ.setdefault("IXMP4_ACCESS_LOG", str(access_file.absolute()))
        os.environ.setdefault("IXMP4_DEBUG_LOG", str(debug_file.absolute()))
        os.environ.setdefault("IXMP4_ERROR_LOG", str(error_file.absolute()))

        logging_config = here / f"logging/{config}.conf"
        logging.config.fileConfig(logging_config, disable_existing_loggers=False)

    def check_credentials(self):
        if self.default_credentials is not None:
            username, password = self.default_credentials
            ManagerAuth(username, password, self.manager_url)



from logging.config import fileConfig

from alembic import context
from sqlalchemy import create_engine

from ixmp4.conf import settings
from ixmp4.data.db import BaseModel

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config
dsn = config.get_main_option("sqlalchemy.url", settings.migration_db_uri)

# Interpret the config file for Python logging.
# This line sets up loggers basically.
assert config.config_file_name is not None
fileConfig(config.config_file_name)

target_metadata = BaseModel.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def include_object(object, name, type_, reflected, compare_to):
    if (
        type_ == "column"
        and not reflected
        and object.info.get("skip_autogenerate", False)
    ):
        return False
    else:
        return True


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        url=dsn,
        target_metadata=target_metadata,
        compare_type=True,
        literal_binds=True,
        render_as_batch=True,
        include_object=include_object,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = create_engine(dsn)

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            compare_type=True,
            render_as_batch=True,
            include_object=include_object,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import pandas as pd

from ..base import BaseFacade
from .variable import VariableRepository


class IamcRepository(BaseFacade):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.variables = VariableRepository(_backend=self.backend)

    def tabulate(self, join_runs=True, **filters) -> pd.DataFrame:
        # return only default runs unless a run-filter is provided
        if "run" not in filters:
            filters["run"] = {"default_only": True}

        df = self.backend.iamc.datapoints.tabulate(
            join_parameters=True,
            join_runs=join_runs,
            **filters,
        ).dropna(how="all", axis="columns")

        if not df.empty:
            df = df.drop(columns=["time_series__id"])
            df.unit = df.unit.replace({"dimensionless": ""})

        return df



from fastapi import APIRouter, Depends, Query, Body


from ixmp4.data.backend import Backend
from ixmp4.data import api
from ixmp4.data.db.model.filter import ModelFilter

from .base import BaseModel
from .decorators import autodoc
from . import deps


router: APIRouter = APIRouter(
    prefix="/models",
    tags=["models"],
)


class ModelInput(BaseModel):
    name: str


class EnumerationOutput(BaseModel):
    __root__: list[api.Model] | api.DataFrame


@autodoc
@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: ModelFilter = Depends(),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.models.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: ModelFilter = Body(None),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.models.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.post("/", response_model=api.Model)
def create(
    model: ModelInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.models.create(**model.dict())



from typing import Iterable, ClassVar
from datetime import datetime
import pandas as pd

from ixmp4.data import abstract
from .docs import Docs, DocsRepository
from . import base


class Unit(base.BaseModel):
    NotFound: ClassVar = abstract.Unit.NotFound
    NotUnique: ClassVar = abstract.Unit.NotUnique
    DeletionPrevented: ClassVar = abstract.Unit.DeletionPrevented

    id: int
    name: str
    created_at: datetime | None
    created_by: str | None


class UnitDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/units/"


class UnitRepository(
    base.Creator[Unit],
    base.Deleter[Unit],
    base.Retriever[Unit],
    base.Enumerator[Unit],
    abstract.UnitRepository,
):
    model_class = Unit
    prefix = "units/"
    enumeration_method = "PATCH"

    def __init__(self, client, *args, **kwargs) -> None:
        super().__init__(client, *args, **kwargs)
        self.docs = UnitDocsRepository(self.client)

    def create(
        self,
        name: str,
    ) -> Unit:
        return super().create(name=name)

    def delete(self, id: int) -> None:
        super().delete(id)

    def get(self, name: str) -> Unit:
        return super().get(name=name)

    def list(self, *args, **kwargs) -> Iterable[Unit]:
        return super().list(*args, **kwargs)

    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    def enumerate(self, *args, **kwargs) -> Iterable[Unit] | pd.DataFrame:
        return super().enumerate(*args, **kwargs)



"""
This module contains sqlalchemy database models and repositories.
"""
# flake8: noqa

from .base import BaseModel, BaseRepository
from .unit import Unit, UnitRepository
from .region import Region, RegionRepository
from .iamc import (
    Variable,
    VariableRepository,
    TimeSeriesRepository,
    TimeSeries,
    Measurand,
    MeasurandRepository,
    DataPoint,
    UniversalDataPoint,
    OracleDataPoint,
    #    AnnualDataPoint,
    #    SubAnnualDataPoint,
    #    CategoricalDataPoint,
    DataPointRepository,
)
from .model import Model, ModelRepository
from .run import Run, RunRepository
from .meta import RunMetaEntry, RunMetaEntryRepository
from .scenario import Scenario, ScenarioRepository
from .docs import docs_model, BaseDocsRepository



from typing import Iterable, ClassVar

from datetime import datetime
import pandas as pd

from ixmp4.data.abstract import Variable as VariableModel, Docs as DocsModel
from ixmp4.core.base import BaseModelFacade, BaseFacade


class Variable(BaseModelFacade):
    _model: VariableModel
    NotFound: ClassVar = VariableModel.NotFound
    NotUnique: ClassVar = VariableModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self):
        try:
            return self.backend.iamc.variables.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description):
        if description is None:
            self.backend.iamc.variables.docs.delete(self.id)
        else:
            self.backend.iamc.variables.docs.set(self.id, description)

    @docs.deleter
    def docs(self):
        try:
            self.backend.iamc.variables.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Variable {self.id} name={self.name}>"


class VariableRepository(BaseFacade):
    def create(
        self,
        name: str,
    ) -> Variable:
        model = self.backend.iamc.variables.create(name)
        return Variable(_backend=self.backend, _model=model)

    def get(self, name: str) -> Variable:
        model = self.backend.iamc.variables.get(name)
        return Variable(_backend=self.backend, _model=model)

    def list(self, name: str | None = None) -> Iterable[Variable]:
        variables = self.backend.iamc.variables.list(name=name)
        return [Variable(_backend=self.backend, _model=v) for v in variables]

    def tabulate(self, name: str | None = None) -> pd.DataFrame:
        return self.backend.iamc.variables.tabulate(name=name)

    def _get_variable_id(self, variable: str) -> int | None:
        if variable is None:
            return None
        elif isinstance(variable, str):
            obj = self.backend.iamc.variables.get(variable)
            return obj.id
        else:
            raise ValueError(f"Invalid reference to variable: {variable}")

    def get_docs(self, name: str) -> str | None:
        variable_id = self._get_variable_id(name)
        if variable_id is None:
            return None
        try:
            return self.backend.iamc.variables.docs.get(
                dimension_id=variable_id
            ).description
        except DocsModel.NotFound:
            return None

    def set_docs(self, name: str, description: str | None) -> str | None:
        if description is None:
            self.delete_docs(name=name)
            return None
        variable_id = self._get_variable_id(name)
        if variable_id is None:
            return None
        return self.backend.iamc.variables.docs.set(
            dimension_id=variable_id, description=description
        ).description

    def delete_docs(self, name: str) -> None:
        # TODO: this function is failing silently, which we should avoid
        variable_id = self._get_variable_id(name)
        if variable_id is None:
            return None
        try:
            self.backend.iamc.variables.docs.delete(dimension_id=variable_id)
            return None
        except DocsModel.NotFound:
            return None



from typing import List

import pandas as pd

from .base import BaseFacade


class MetaRepository(BaseFacade):
    def tabulate(self, run_ids: List[int]) -> pd.DataFrame:
        # TODO: accept list of `Run` instances as arg
        # TODO: expand run-id to model-scenario-version-id columns
        return (
            self.backend.meta.tabulate(run_ids=run_ids)
            .drop(columns=["id", "type"])
            .rename(columns={"run__id": "run_id"})
        )



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import abstract, types

from ixmp4.data.db.model.model import Model
from ixmp4.data.db.scenario.model import Scenario
from .. import base


class Run(base.BaseModel):
    NotFound: ClassVar = abstract.Run.NotFound
    NotUnique: ClassVar = abstract.Run.NotUnique
    DeletionPrevented: ClassVar = abstract.Run.DeletionPrevented

    NoDefaultVersion: ClassVar = abstract.Run.NoDefaultVersion

    __table_args__ = (db.UniqueConstraint("model__id", "scenario__id", "version"),)

    model__id: types.Integer = db.Column(
        db.ForeignKey("model.id"), nullable=False, index=True
    )
    model: types.Mapped[Model] = db.relationship(
        "Model",
        backref="run",
        foreign_keys=[model__id],
    )

    scenario__id: types.Integer = db.Column(
        db.ForeignKey("scenario.id"),
        nullable=False,
        index=True,
    )
    scenario: types.Mapped[Scenario] = db.relationship(
        "Scenario",
        backref="run",
        foreign_keys=[scenario__id],
    )

    version: types.Integer = db.Column(db.Integer, nullable=False)
    is_default: types.Boolean = db.Column(db.Boolean, default=False, nullable=False)



import logging
from typing import Callable, AsyncGenerator

import jwt
from fastapi import Path, Header, Depends

from ixmp4.data.backend.db import SqlAlchemyBackend
from ixmp4.conf import settings
from ixmp4.conf.user import local_user, anonymous_user, User
from ixmp4.conf.manager import ManagerConfig
from ixmp4.conf.auth import SelfSignedAuth
from ixmp4.core.exceptions import (
    InvalidToken,
    Forbidden,
    PlatformNotFound,
)

logger = logging.getLogger(__name__)
manager = ManagerConfig(settings.manager_url, SelfSignedAuth(settings.secret_hs256))


async def validate_token(authorization: str = Header(None)) -> dict | None:
    """Validates a JSON Web Token with the secret supplied in the
    `IXMP4_SECRET_HS256` environment variable."""

    if authorization is None:
        return None  # anonymous user

    encoded_jwt = authorization.split(" ")[1]
    try:
        return jwt.decode(
            encoded_jwt, settings.secret_hs256, leeway=300, algorithms=["HS256"]
        )
    except jwt.InvalidTokenError as e:
        raise InvalidToken("The supplied token is expired or invalid.") from e


async def do_not_validate_token(authorization: str = Header(None)) -> dict | None:
    """Override dependency used for skipping authentication while testing."""
    return {"user": local_user.dict()}


async def get_user(token: dict | None = Depends(validate_token)) -> User:
    """Returns a user object for permission checks."""
    if token is None:
        return anonymous_user

    try:
        user_dict = token["user"]
        user_dict["jti"] = token.get("jti", None)
    except KeyError as e:
        raise InvalidToken("The supplied token is malformed.") from e

    return User(**user_dict)


async def get_version():
    from ixmp4 import __version__

    return __version__


async def get_managed_backend(
    platform: str = Path("default"), user: User = Depends(get_user)
) -> AsyncGenerator[SqlAlchemyBackend, None]:
    """Returns a platform backend for a platform name as a path parameter.
    Also checks user access permissions if in managed mode."""
    info = manager.get_platform(platform, jti=manager.auth.get_user().jti)

    if info.dsn.startswith("http"):
        raise PlatformNotFound(f"Platform '{platform}' was not found.")
    else:
        backend = SqlAlchemyBackend(info)
        with backend.auth(user, manager, info) as auth:
            if auth.is_accessible:
                yield backend
            else:
                raise Forbidden(
                    f"Access to platform '{platform}' denied "
                    "due to insufficient permissions."
                )
        backend.close()


async def get_toml_backend(
    platform: str = Path("default"), user: User = Depends(get_user)
) -> AsyncGenerator[SqlAlchemyBackend, None]:
    logger.debug("Looking for platform in `platforms.toml`.")
    info = settings.toml.get_platform(platform)
    if info.dsn.startswith("http"):
        raise PlatformNotFound(f"Platform '{platform}' was not found.")
    else:
        backend = SqlAlchemyBackend(info)
        yield backend
        backend.close()


if settings.managed:
    get_backend = get_managed_backend
else:
    get_backend = get_toml_backend


def get_test_backend_dependency(backend, auth_params) -> Callable:
    async def get_memory_backend(
        platform: str = Path("default"), user: User = Depends(get_user)
    ) -> AsyncGenerator[SqlAlchemyBackend, None]:
        """Override dependency which always yields a test backend."""
        with backend.auth(*auth_params, overlap_ok=True):
            yield backend

    return get_memory_backend



from typing import Optional, Iterable, Union, ClassVar, cast

import pandera as pa
import pandas as pd
from pandera.typing import DataFrame, Series

from sqlalchemy.exc import NoResultFound

from ixmp4 import db
from ixmp4.data import abstract, types
from ixmp4.data.db.run import Run
from ixmp4.data.db.model import Model
from ixmp4.core.exceptions import (
    InvalidRunMeta,
)
from ixmp4.core.decorators import check_types

from ..auth.decorators import guard
from . import base


class RemoveRunMetaEntryFrameSchema(pa.SchemaModel):
    key: Series[pa.String] = pa.Field(coerce=True)
    run__id: Series[pa.Int] = pa.Field(coerce=True)

    class Config:
        strict = True
        coerce = True


class AddRunMetaEntryFrameSchema(RemoveRunMetaEntryFrameSchema):
    value: Series[pa.Object] = pa.Field(coerce=True)


class UpdateRunMetaEntryFrameSchema(AddRunMetaEntryFrameSchema):
    id: Series[pa.Int] = pa.Field(coerce=True)


class RunMetaEntry(base.BaseModel):
    NotFound: ClassVar = abstract.RunMetaEntry.NotFound
    NotUnique: ClassVar = abstract.RunMetaEntry.NotUnique
    DeletionPrevented: ClassVar = abstract.RunMetaEntry.DeletionPrevented

    Type: ClassVar = abstract.RunMetaEntry.Type

    _column_map = {
        abstract.RunMetaEntry.Type.INT: "value_int",
        abstract.RunMetaEntry.Type.STR: "value_str",
        abstract.RunMetaEntry.Type.FLOAT: "value_float",
        abstract.RunMetaEntry.Type.BOOL: "value_bool",
    }

    __table_args__ = (
        db.UniqueConstraint(
            "run__id",
            "key",
        ),
    )
    updateable_columns = [
        "type",
        "value_int",
        "value_str",
        "value_float",
        "value_bool",
    ]

    run__id: types.Integer = db.Column(
        db.Integer,
        db.ForeignKey("run.id"),
        nullable=False,
        index=True,
    )
    run = db.relationship(
        "Run",
        backref="meta",
        foreign_keys=[run__id],
    )

    key: types.String = db.Column(db.String(1023), nullable=False)
    type: types.String = db.Column(db.String(20), nullable=False)

    value_int: types.Integer = db.Column(db.Integer, nullable=True)
    value_str: types.String = db.Column(db.String(1023), nullable=True)
    value_float: types.Float = db.Column(db.Float, nullable=True)
    value_bool: types.Boolean = db.Column(db.Boolean, nullable=True)

    @property
    def value(self) -> abstract.MetaValue:
        type_ = RunMetaEntry.Type(self.type)
        col = self._column_map[type_]
        return getattr(self, col)

    def __init__(self, *args, **kwargs) -> None:
        value = kwargs.pop("value")
        value_type = type(value)
        try:
            type_ = RunMetaEntry.Type.from_pytype(value_type)
            col = self._column_map[type_]
        except KeyError:
            raise InvalidRunMeta(
                f"Invalid type `{value_type}` for value of `RunMetaEntry`."
            )
        kwargs["type"] = type_
        kwargs[col] = value
        super().__init__(*args, **kwargs)


class RunMetaEntryRepository(
    base.Creator[RunMetaEntry],
    base.Enumerator[RunMetaEntry],
    base.BulkUpserter[RunMetaEntry],
    base.BulkDeleter[RunMetaEntry],
    abstract.RunMetaEntryRepository,
):
    model_class = RunMetaEntry

    def add(
        self, run__id: int, key: str, value: Union[str, int, bool, float]
    ) -> RunMetaEntry:
        if self.backend.auth_context is not None:
            self.backend.runs.check_access(
                {run__id},
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        entry = RunMetaEntry(run__id=run__id, key=key, value=value)
        self.session.add(entry)
        return entry

    def check_df_access(self, df: pd.DataFrame):
        if self.backend.auth_context is not None:
            ts_ids = set(df["run__id"].unique().tolist())
            self.backend.runs.check_access(
                ts_ids,
                access_type="edit",
                is_default=None,
                default_only=False,
            )

    @guard("edit")
    def create(self, *args, **kwargs) -> RunMetaEntry:
        return super().create(*args, **kwargs)

    @guard("view")
    def get(self, run__id: int, key: str) -> RunMetaEntry:
        if self.backend.auth_context is not None:
            self.backend.runs.check_access(
                {run__id},
                access_type="view",
                is_default=None,
                default_only=False,
            )

        exc = self.select(
            run_ids=[run__id],
            keys=[key],
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise RunMetaEntry.NotFound(
                run__id=run__id,
                key=key,
            )

    @guard("edit")
    def delete(self, id: int) -> None:
        if self.backend.auth_context is not None:
            try:
                pre_exc = db.select(RunMetaEntry).where(RunMetaEntry.id == id)
                meta = self.session.execute(pre_exc).scalar_one()
            except NoResultFound:
                raise RunMetaEntry.NotFound(
                    id=id,
                )
            self.backend.runs.check_access(
                {meta.run__id},
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        exc = db.delete(RunMetaEntry).where(RunMetaEntry.id == id)

        try:
            self.session.execute(exc)
            self.session.commit()
        except NoResultFound:
            raise RunMetaEntry.NotFound(
                id=id,
            )

    def join_auth(self, exc: db.sql.Select) -> db.sql.Select:
        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, RunMetaEntry.run)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return super().join_auth(exc)

    def select(
        self,
        *,
        run_ids: Optional[Iterable[int]] = None,
        keys: Optional[Iterable[str]] = None,
        _access_type: str = "view",
    ) -> db.sql.Select:
        exc: db.sql.Select = db.select(RunMetaEntry)
        exc = self.apply_auth(exc, _access_type)

        if run_ids is not None:
            exc = exc.where(RunMetaEntry.run__id.in_(run_ids))

        if keys is not None:
            exc = exc.where(RunMetaEntry.key.in_(keys))

        return exc

    @guard("view")
    def list(self, *args, **kwargs) -> Iterable[RunMetaEntry]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(
        self,
        *args,
        _raw: bool | None = False,
        **kwargs,
    ) -> pd.DataFrame:
        if _raw:
            return super().tabulate(*args, **kwargs)
        df = super().tabulate(*args, **kwargs)

        if df.empty:
            return pd.DataFrame([], columns=["id", "run__id", "type", "key", "value"])

        def map_value_column(df: pd.DataFrame):
            type_str = df.name
            type_ = RunMetaEntry.Type(type_str)
            col = RunMetaEntry._column_map[type_]
            df["value"] = df[col]
            return df.drop(columns=RunMetaEntry._column_map.values())

        return df.groupby("type", group_keys=False).apply(map_value_column)

    @check_types
    @guard("edit")
    def bulk_upsert(self, df: DataFrame[AddRunMetaEntryFrameSchema]) -> None:
        self.check_df_access(df)
        df["type"] = df["value"].map(type).map(RunMetaEntry.Type.from_pytype)

        for type_, type_df in df.groupby("type"):
            type_ = cast(RunMetaEntry.Type, type_)
            col = RunMetaEntry._column_map[type_]
            null_cols = set(RunMetaEntry._column_map.values()) - set([col])
            type_df["type"] = type_df["type"].map(lambda x: x.value)
            type_df = type_df.rename(columns={"value": col})

            # ensure all other columns are overwritten
            for nc in null_cols:
                type_df[nc] = None

            super().bulk_upsert(type_df)

    @check_types
    @guard("edit")
    def bulk_delete(self, df: DataFrame[RemoveRunMetaEntryFrameSchema]) -> None:
        self.check_df_access(df)
        return super().bulk_delete(df)



from fastapi import APIRouter, Depends, Query, Body

from ixmp4.core.exceptions import BadRequest
from ixmp4.data.backend import Backend
from ixmp4.data import api
from ixmp4.data.db.iamc.datapoint.filter import DataPointFilter
from ..decorators import autodoc

from .. import deps
from ..base import BaseModel


router: APIRouter = APIRouter(
    prefix="/datapoints",
    tags=["iamc", "datapoints"],
)


class EnumerationOutput(BaseModel):
    __root__: list[api.DataPoint] | api.DataFrame


@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: DataPointFilter = Depends(),
    join_parameters: bool | None = Query(False),
    join_runs: bool = Query(False),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    if (join_parameters or join_runs) and not table:
        raise BadRequest(
            "`join_parameters` or `join_run` can only be used with `table=true`."
        )

    return backend.iamc.datapoints.enumerate(
        table=bool(table),
        _filter=filter,
        join_parameters=join_parameters,
        join_runs=join_runs,
    )


@autodoc
@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: DataPointFilter = Body(DataPointFilter()),
    join_parameters: bool | None = Query(False),
    join_runs: bool = Query(False),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    """This endpoint is used to retrieve and optionally filter data.add()

    Filter parameters are provided as keyword arguments.

    The available filters can be found here:
    :class:`ixmp4.data.db.iamc.datapoint.filter.DataPointFilter`.

    Examples
    --------

    Filter data points for a given model, scenario combination, and a number of years:

    ..  code-block:: json

        {
            "model" : {"name": "model 1"},
            "scenario" : {"name": "scenario 1"},
            "year__in" : [2020, 2025]
        }

    Return all data for a given variable:

    .. code-block:: json

        {
            "variable": {"name": "Final Energy"}
        }


    """
    if (join_parameters or join_runs) and not table:
        raise BadRequest(
            "`join_parameters` or `join_run` can only be used with `table=true`."
        )

    return backend.iamc.datapoints.enumerate(
        table=bool(table),
        _filter=filter,
        join_parameters=join_parameters,
        join_runs=join_runs,
    )


@autodoc
@router.post("/bulk/")
def bulk_upsert(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.datapoints.bulk_upsert(df.to_pandas())


@autodoc
@router.patch("/bulk/")
def bulk_delete(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.datapoints.bulk_delete(df.to_pandas())



from fastapi import APIRouter, Depends, Query, Body, Path

from ixmp4.data.backend import Backend
from ixmp4.data import api
from ixmp4.data.db.region.filter import RegionFilter

from .base import BaseModel
from .decorators import autodoc
from . import deps

router: APIRouter = APIRouter(
    prefix="/regions",
    tags=["regions"],
)


class RegionInput(BaseModel):
    name: str
    hierarchy: str


class EnumerationOutput(BaseModel):
    __root__: list[api.Region] | api.DataFrame


@autodoc
@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: RegionFilter = Depends(),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.regions.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: RegionFilter = Body(None),
    table: bool = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.regions.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.post("/", response_model=api.Region)
def create(
    region: RegionInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.regions.create(**region.dict())


@router.delete("/{id}/")
def delete(
    id: int = Path(None),
    backend: Backend = Depends(deps.get_backend),
):
    backend.regions.delete(id)



# flake8: noqa

from .timeseries import TimeSeries, TimeSeriesRepository
from .datapoint import (
    DataPoint,
    #    AnnualDataPoint,
    #    SubAnnualDataPoint,
    #    CategoricalDataPoint,
    DataPointRepository,
)
from .variable import Variable, VariableRepository



from ixmp4.conf.base import PlatformInfo
from ixmp4.data.abstract import (
    DataPointRepository,
    RunRepository,
    RunMetaEntryRepository,
    TimeSeriesRepository,
    UnitRepository,
    RegionRepository,
    ScenarioRepository,
    ModelRepository,
    VariableRepository,
)


class IamcSubobject(object):
    datapoints: DataPointRepository
    timeseries: TimeSeriesRepository
    variables: VariableRepository


class Backend(object):
    info: PlatformInfo
    runs: RunRepository
    meta: RunMetaEntryRepository
    regions: RegionRepository
    units: UnitRepository
    scenarios: ScenarioRepository
    models: ModelRepository
    iamc: IamcSubobject

    def __init__(self, info: PlatformInfo) -> None:
        self.info = info
        self.iamc = IamcSubobject()



from typing import Any, Mapping, Iterable

import pandas as pd
import numpy as np

from sqlalchemy import select
from sqlalchemy.orm import Bundle

from ixmp4 import db
from ixmp4.db import utils
from ixmp4.data import types, abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.iamc.measurand import Measurand

from ..timeseries import (
    TimeSeries as BaseTimeSeries,
    TimeSeriesRepository as BaseTimeSeriesRepository,
)
from ..utils import map_existing

from . import base
from .measurand import MeasurandRepository
from .variable import Variable
from ..run import RunRepository
from ..unit import Unit, UnitRepository
from ..region import Region, RegionRepository


class TimeSeries(BaseTimeSeries, base.BaseModel):
    __table_args__ = (db.UniqueConstraint("run__id", "region__id", "measurand__id"),)

    region__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("region.id"), nullable=False, index=True
    )
    region: types.Mapped[Region] = db.relationship(
        "Region", backref="metadata", foreign_keys=[region__id], lazy="select"
    )

    measurand__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("iamc_measurand.id"), nullable=False, index=True
    )
    measurand: types.Mapped[Measurand] = db.relationship(
        "Measurand", backref="metadata", foreign_keys=[measurand__id], lazy="select"
    )

    @property
    def parameters(self) -> Mapping:
        return {
            "region": self.region.name,
            "unit": self.measurand.unit.name,
            "variable": self.measurand.variable.name,
        }


class TimeSeriesRepository(
    BaseTimeSeriesRepository[TimeSeries], abstract.TimeSeriesRepository
):
    model_class = TimeSeries

    regions: RegionRepository
    measurands: MeasurandRepository
    units: UnitRepository

    def __init__(self, *args, **kwargs) -> None:
        self.runs = RunRepository(*args, **kwargs)
        self.regions = RegionRepository(*args, **kwargs)
        self.measurands = MeasurandRepository(*args, **kwargs)
        self.units = UnitRepository(*args, **kwargs)
        super().__init__(*args, **kwargs)

    @guard("view")
    def get(self, run_id: int, **kwargs: Any) -> TimeSeries:
        return super().get(run_id, **kwargs)

    def filter_by_parameters(
        self, exc: db.sql.Select, parameters: Any
    ) -> db.sql.Select:
        if not utils.is_joined(exc, Measurand):
            exc = exc.join(TimeSeries.measurand)

        for key, col in [
            ("region", TimeSeries.region__id),
            ("variable", Measurand.variable__id),
            ("unit", Measurand.unit__id),
        ]:
            value = parameters.pop(key, None)
            if value is not None:
                exc = exc.where(col == value)

        if len(parameters) > 0:
            raise ValueError(
                "Invalid `parameters` supplied: " + ", ".join(parameters.keys())
            )
        return exc

    def select_joined_parameters(self):
        return (
            select(
                self.bundle,
                Bundle(
                    "Region",
                    Region.name.label("region"),
                ),
                Bundle(
                    "Unit",
                    Unit.name.label("unit"),
                ),
                Bundle(
                    "Variable",
                    Variable.name.label("variable"),
                ),
            )
            .join(Region, onclause=TimeSeries.region__id == Region.id)
            .join(Measurand, onclause=TimeSeries.measurand__id == Measurand.id)
            .join(Unit, onclause=Measurand.unit__id == Unit.id)
            .join(Variable, onclause=Measurand.variable__id == Variable.id)
        )

    @guard("view")
    def list(self, *args, **kwargs) -> Iterable[TimeSeries]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    @guard("edit")
    def bulk_upsert(self, df: pd.DataFrame, create_related: bool = False) -> None:
        if self.backend.auth_context is not None:
            run_ids = set(df["run__id"].unique().tolist())
            self.runs.check_access(
                run_ids,
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        if create_related:
            df = self.map_relationships(df)
            df = df.drop_duplicates()
        super().bulk_upsert(df)

    def map_regions(self, df: pd.DataFrame):
        existing_regions = self.regions.tabulate(name__in=df["region"].unique())
        df, missing = map_existing(
            df,
            existing_df=existing_regions,
            join_on=("name", "region"),
            map=("id", "region__id"),
        )
        if len(missing) > 0:
            raise Region.NotFound(", ".join(missing))

        return df

    def map_measurands(self, df: pd.DataFrame) -> pd.DataFrame:
        df, missing = map_existing(
            df,
            existing_df=self.units.tabulate(),
            join_on=("name", "unit"),
            map=("id", "unit__id"),
        )
        if len(missing) > 0:
            raise Unit.NotFound(", ".join(missing))

        df["measurand__id"] = np.nan

        def map_measurand(df):
            variable_name, unit__id = df.name
            measurand = self.measurands.get_or_create(
                variable_name=variable_name, unit__id=int(unit__id)
            )
            df["measurand__id"] = measurand.id
            return df

        return pd.DataFrame(
            df.groupby(["variable", "unit__id"], group_keys=False).apply(map_measurand)
        )

    def map_relationships(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.map_regions(df)
        df = self.map_measurands(df)
        return df



from ixmp4.db import filters, utils

from . import Run, Model, Scenario, Region, Unit, Variable, Measurand, TimeSeries


class ModelFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    name: filters.String

    class Config:
        sqla_model = Model

    def join(self, exc, **kwargs):
        return exc.join(Model, Run.model)


class ScenarioFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    name: filters.String

    class Config:
        sqla_model = Scenario

    def join(self, exc, **kwargs):
        return exc.join(Scenario, Run.scenario)


class RunFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    version: filters.Integer
    default_only: filters.Boolean = filters.Field(True)
    is_default: filters.Boolean
    model: ModelFilter | None
    scenario: ScenarioFilter | None

    class Config:
        sqla_model = Run

    def filter_default_only(self, exc, c, v, **kwargs):
        if v:
            return exc.where(Run.is_default)
        else:
            return exc

    def join(self, exc, **kwargs):
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, TimeSeries.run)
        return exc


class VariableFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    name: filters.String

    class Config:
        sqla_model = Variable

    def join(self, exc, **kwargs):
        if not utils.is_joined(exc, Measurand):
            exc = exc.join(Measurand, TimeSeries.measurand)
        exc = exc.join(Variable, Measurand.variable)
        return exc


class UnitFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    name: filters.String

    class Config:
        sqla_model = Unit

    def join(self, exc, **kwargs):
        if not utils.is_joined(exc, Measurand):
            exc = exc.join(Measurand, TimeSeries.measurand)
        exc = exc.join(Unit, Measurand.unit)
        return exc


class RegionFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    name: filters.String
    hierarchy: filters.String

    class Config:
        sqla_model = Region

    def join(self, exc, **kwargs):
        exc = exc.join(Region, TimeSeries.region)
        return exc



from fastapi import APIRouter, Depends, Query, Body

from ixmp4.data.backend import Backend
from ixmp4.data import api
from ixmp4.data.db.scenario.filter import ScenarioFilter

from .base import BaseModel
from .decorators import autodoc
from . import deps


router: APIRouter = APIRouter(
    prefix="/scenarios",
    tags=["scenarios"],
)


class ScenarioInput(BaseModel):
    name: str


class EnumerationOutput(BaseModel):
    __root__: list[api.Scenario] | api.DataFrame


@autodoc
@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: ScenarioFilter = Depends(),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.scenarios.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: ScenarioFilter = Body(None),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.scenarios.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.post("/", response_model=api.Scenario)
def create(
    scenario: ScenarioInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.scenarios.create(**scenario.dict())



from fastapi import APIRouter, Depends, Query, Body

from ixmp4.data.backend import Backend
from ixmp4.data.db.model.filter import IamcModelFilter
from ixmp4.data import api

from ..base import BaseModel
from .. import deps

router: APIRouter = APIRouter(
    prefix="/models",
    tags=["iamc", "models"],
)


class EnumerationOutput(BaseModel):
    __root__: list[api.Model] | api.DataFrame


@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: IamcModelFilter = Depends(),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.models.enumerate(table=bool(table), _filter=filter)


@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: IamcModelFilter = Body(IamcModelFilter()),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.models.enumerate(table=bool(table), _filter=filter)



from typing import Iterable, Optional, Union
from datetime import datetime
import pandas as pd

from ixmp4.data.abstract import Region as RegionModel, Docs as DocsModel
from .base import BaseModelFacade, BaseFacade


class Region(BaseModelFacade):
    _model: RegionModel
    NotUnique = RegionModel.NotUnique
    NotFound = RegionModel.NotFound
    DeletionPrevented = RegionModel.DeletionPrevented

    @property
    def id(self) -> int:
        """Unique id."""
        return self._model.id

    @property
    def name(self) -> str:
        """Region name."""
        return self._model.name

    @property
    def hierarchy(self) -> str:
        """Region hierarchy."""
        return self._model.hierarchy

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    def delete(self):
        """Deletes the region from the database."""
        self.backend.regions.delete(self._model.id)

    @property
    def docs(self):
        try:
            return self.backend.regions.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description):
        if description is None:
            self.backend.regions.docs.delete(self.id)
        else:
            self.backend.regions.docs.set(self.id, description)

    @docs.deleter
    def docs(self):
        try:
            self.backend.regions.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Region {self.id} name={self.name}>"


class RegionRepository(BaseFacade):
    def _get_region_id(self, region: Optional[Union[str, int, "Region"]]) -> int | None:
        if region is None:
            return None
        elif isinstance(region, str):
            obj = self.backend.regions.get(region)
            return obj.id
        elif isinstance(region, int):
            return region
        elif isinstance(region, Region):
            return region.id
        else:
            raise ValueError(f"Invalid reference to region: {region}")

    def create(
        self,
        name: str,
        hierarchy: str,
    ) -> Region:
        model = self.backend.regions.create(name, hierarchy)
        return Region(_backend=self.backend, _model=model)

    def get(self, name: str) -> Region:
        model = self.backend.regions.get(name)
        return Region(_backend=self.backend, _model=model)

    def delete(self, x: Region | int | str):
        if isinstance(x, Region):
            id = x.id
        elif isinstance(x, int):
            id = x
        elif isinstance(x, str):
            model = self.backend.regions.get(x)
            id = model.id
        else:
            raise TypeError("Invalid argument: Must be `Region`, `int` or `str`.")

        self.backend.regions.delete(id)

    def list(
        self,
        name: str | None = None,
        hierarchy: str | None = None,
    ) -> Iterable[Region]:
        regions = self.backend.regions.list(name=name, hierarchy=hierarchy)
        return [Region(_backend=self.backend, _model=r) for r in regions]

    def tabulate(
        self,
        name: str | None = None,
        hierarchy: str | None = None,
    ) -> pd.DataFrame:
        return self.backend.regions.tabulate(name=name, hierarchy=hierarchy)

    def get_docs(self, name: str) -> str | None:
        region_id = self._get_region_id(name)
        if region_id is None:
            return None
        try:
            return self.backend.regions.docs.get(dimension_id=region_id).description
        except DocsModel.NotFound:
            return None

    def set_docs(self, name: str, description: str | None) -> str | None:
        if description is None:
            self.delete_docs(name=name)
            return None
        region_id = self._get_region_id(name)
        if region_id is None:
            return None
        return self.backend.regions.docs.set(
            dimension_id=region_id, description=description
        ).description

    def delete_docs(self, name: str) -> None:
        # TODO: this function is failing silently, which we should avoid
        region_id = self._get_region_id(name)
        if region_id is None:
            return None
        try:
            self.backend.regions.docs.delete(dimension_id=region_id)
            return None
        except DocsModel.NotFound:
            return None



# flake8: noqa

from .base import BaseModel, DataFrame

from .run import Run, RunRepository
from .meta import RunMetaEntry, RunMetaEntryRepository
from .unit import Unit, UnitRepository
from .region import Region, RegionParent, RegionRepository
from .scenario import Scenario, ScenarioRepository
from .model import Model, ModelRepository
from .docs import Docs, DocsRepository

from .iamc import (
    Variable,
    VariableRepository,
    #     Measurand,
    #     MeasurandRepository,
    TimeSeries,
    TimeSeriesRepository,
    DataPoint,
    #    AnnualDataPoint,
    #    SubAnnualDataPoint,
    #    CategoricalDataPoint,
    DataPointRepository,
)



from typing import Optional

import pandas as pd
import pandera as pa
from pandera.typing import Series

from ixmp4.data.abstract import DataPoint as DataPointModel

from ..base import BaseFacade
from .repository import IamcRepository
from ..run import RunModel as Run
from ..utils import substitute_type


def to_dimensionless(df: pd.DataFrame) -> pd.DataFrame:
    if "dimensionless" in df.unit:
        raise ValueError(
            "Unit name 'dimensionless' is reserved, use an empty string '' instead."
        )
    df.unit.replace({"": "dimensionless"}, inplace=True)
    return df


class RemoveDataPointFrameSchema(pa.SchemaModel):
    type: Optional[Series[pa.String]] = pa.Field(isin=[t for t in DataPointModel.Type])
    step_year: Optional[Series[pa.Int]] = pa.Field(coerce=True, nullable=True)
    step_datetime: Optional[Series[pa.DateTime]] = pa.Field(coerce=True, nullable=True)
    step_category: Optional[Series[pa.String]] = pa.Field(nullable=True)

    region: Optional[Series[pa.String]] = pa.Field(coerce=True)
    unit: Optional[Series[pa.String]] = pa.Field(coerce=True)
    variable: Optional[Series[pa.String]] = pa.Field(coerce=True)


class AddDataPointFrameSchema(RemoveDataPointFrameSchema):
    value: Series[pa.Float] = pa.Field(coerce=True)


class IamcData(BaseFacade):
    """IAMC data.

    Parameters
    ----------
    backend : ixmp4.data.backend.Backend
        Data source backend.
    run : ixmp4.base.run.Run
        Model run.
    """

    run: Run

    def __init__(self, *args, run: Run, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.run = run
        self.repository = IamcRepository(_backend=self.backend)

    def _contract_parameters(self, df: pd.DataFrame) -> pd.DataFrame:
        ts_df = df[["region", "variable", "unit", "run__id"]].drop_duplicates()
        self.backend.iamc.timeseries.bulk_upsert(ts_df, create_related=True)

        ts_df = self.backend.iamc.timeseries.tabulate(
            run_ids=[self.run.id], join_parameters=True
        )
        ts_df = ts_df.rename(columns={"id": "time_series__id"})

        return pd.merge(
            df,
            ts_df,
            how="left",
            on=["run__id", "region", "unit", "variable"],
            suffixes=(None, "_y"),
        )

    def add(
        self,
        df: pd.DataFrame,
        type: Optional[DataPointModel.Type] = None,
    ):
        df = AddDataPointFrameSchema.validate(df)  # type:ignore
        df = to_dimensionless(df.copy())
        df["run__id"] = self.run.id
        df = self._contract_parameters(df)
        substitute_type(df, type)
        self.backend.iamc.datapoints.bulk_upsert(df)

    def remove(
        self,
        df: pd.DataFrame,
        type: Optional[DataPointModel.Type] = None,
    ):
        df = RemoveDataPointFrameSchema.validate(df)  # type:ignore
        df = to_dimensionless(df.copy())
        df["run__id"] = self.run.id
        df = self._contract_parameters(df)
        substitute_type(df, type)
        df = df.drop(columns=["unit", "variable", "region"])
        self.backend.iamc.datapoints.bulk_delete(df)

    def tabulate(self, **filters) -> pd.DataFrame:
        # these filters do not make sense when applied from a Run
        illegal_filters = [i for i in filters if i in ["run", "model", "scenario"]]
        if illegal_filters:
            raise ValueError(
                f"Illegal filter for `Run.iamc.tabulate()`: {illegal_filters}"
            )

        return self.repository.tabulate(
            run={"id": self.run.id}, join_runs=False, **filters
        )



from typing import Iterable, ClassVar
from collections import UserDict

import pandas as pd

from ixmp4.data.abstract import Run as RunModel
from .base import BaseModelFacade, BaseFacade
from .iamc import IamcData


class Run(BaseModelFacade):
    _model: RunModel
    _meta: "RunMetaFacade"
    NoDefaultVersion: ClassVar = RunModel.NoDefaultVersion
    NotFound: ClassVar = RunModel.NotFound
    NotUnique: ClassVar = RunModel.NotUnique

    def __init__(
        self,
        model: str | None = None,
        scenario: str | None = None,
        version: int | str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if getattr(self, "_model", None) is None:
            if (model is None) or (scenario is None):
                raise TypeError("`Run` requires `model` and `scenario`")

            if version is None:
                self._model = self.backend.runs.get_default_version(model, scenario)
            elif version == "new":
                self._model = self.backend.runs.create(model, scenario)
            elif isinstance(version, int):
                self._model = self.backend.runs.get(model, scenario, version)
            else:
                raise ValueError(
                    "Invalid value for `version`, must be 'new' or integer."
                )
            self.version = self._model.version

        self.iamc = IamcData(_backend=self.backend, run=self._model)
        self._meta = RunMetaFacade(_backend=self.backend, run=self._model)

    @property
    def model(self):
        """Associated model."""
        return self._model.model

    @property
    def scenario(self):
        """Associated scenario."""
        return self._model.scenario

    @property
    def id(self):
        """Unique id."""
        return self._model.id

    @property
    def meta(self):
        "Meta indicator data (`dict`-like)."
        return self._meta

    @meta.setter
    def meta(self, meta):
        self._meta._set(meta)

    def set_as_default(self):
        """Sets this run as the default version for its `model` + `scenario` combination."""
        self.backend.runs.set_as_default_version(self._model.id)

    def unset_as_default(self):
        """Unsets this run as the default version."""
        self.backend.runs.unset_as_default_version(self._model.id)


class RunRepository(BaseFacade):
    def list(self, default_only: bool = True) -> Iterable[Run]:
        return [
            Run(_backend=self.backend, _model=r)
            for r in self.backend.runs.list(default_only=default_only)
        ]

    def tabulate(self, default_only: bool = True) -> pd.DataFrame:
        runs = self.backend.runs.tabulate(default_only=default_only)
        runs["model"] = runs["model__id"].map(
            dict([(m.id, m.name) for m in self.backend.models.list()])
        )
        runs["scenario"] = runs["scenario__id"].map(
            dict([(s.id, s.name) for s in self.backend.scenarios.list()])
        )
        return runs[["id", "model", "scenario", "version", "is_default"]]


class RunMetaFacade(BaseFacade, UserDict):
    run: RunModel

    def __init__(self, run: RunModel, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.run = run
        self.df, self.data = self._get()

    def _get(self) -> tuple[pd.DataFrame, dict]:
        df = self.backend.meta.tabulate(run_ids=[self.run.id])
        if df.empty:
            return df, {}
        return df, dict(zip(df["key"], df["value"]))

    def _set(self, meta: dict):
        df = pd.DataFrame({"key": self.data.keys()})
        df["run__id"] = self.run.id
        self.backend.meta.bulk_delete(df)
        df = pd.DataFrame({"key": meta.keys(), "value": meta.values()})
        df["run__id"] = self.run.id
        self.backend.meta.bulk_upsert(df)
        self.df, self.data = self._get()

    def __setitem__(self, key, value: int | float | str | bool):
        try:
            del self[key]
        except KeyError:
            pass

        self.backend.meta.create(self.run.id, key, value)
        self.df, self.data = self._get()

    def __delitem__(self, key):
        id = dict(zip(self.df["key"], self.df["id"]))[key]
        self.backend.meta.delete(id)
        self.df, self.data = self._get()



from typing import (
    Mapping,
    Sequence,
    Iterable,
    Type,
    ClassVar,
    TypeVar,
    Generic,
    Any,
)
from json.decoder import JSONDecodeError

import httpx
import pandas as pd
from pydantic import BaseModel as PydanticBaseModel

from ixmp4.core.exceptions import (
    registry,
    ImproperlyConfigured,
    UnknownApiError,
    IxmpError,
)


class BaseModel(PydanticBaseModel):
    NotFound: ClassVar[type[IxmpError]]
    NotUnique: ClassVar[type[IxmpError]]

    class Config:
        orm_mode = True


class DataFrame(PydanticBaseModel):
    index: list | None
    columns: list[str]
    dtypes: list[str]
    data: list

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, df: pd.DataFrame | dict):
        if isinstance(df, pd.DataFrame):
            return cls.from_pandas(df)
        else:
            return cls(**df)

    class Config:
        json_encoders = {pd.Timestamp: lambda x: x.isoformat()}

    @classmethod
    def from_pandas(cls, df: pd.DataFrame) -> "DataFrame":
        columns = []
        dtypes = []
        for c in df.columns:
            columns.append(c)
            dtypes.append(df[c].dtype.name)

        return DataFrame(
            index=df.index.tolist(),
            columns=columns,
            dtypes=dtypes,
            data=df.values.tolist(),
        )

    def to_pandas(self) -> pd.DataFrame:
        df = pd.DataFrame(
            index=self.index or None,
            columns=self.columns,
            data=self.data,
        )
        for c, dt in zip(self.columns, self.dtypes):
            # there seems to be a type incompatbility between StrDtypeArg and str
            df[c] = df[c].astype(dt)  # type: ignore
        return df


ModelType = TypeVar("ModelType", bound=BaseModel)


class BaseRepository(Generic[ModelType]):
    model_class: Type[ModelType]
    prefix: ClassVar[str]
    enumeration_method: str = "GET"

    client: httpx.Client

    def __init__(self, client: httpx.Client, *args, **kwargs) -> None:
        self.client = client

    def sanitize_params(self, params: dict):
        return {k: params[k] for k in params if params[k] is not None}

    def get_remote_exception(self, res: httpx.Response, status_code: int):
        try:
            json = res.json()
        except (ValueError, JSONDecodeError):
            raise UnknownApiError(content=res.text, response_status=res.status_code)

        try:
            error_name = json["error_name"]
        except KeyError:
            raise UnknownApiError(json_data=json, response_status=status_code)

        try:
            exc = registry[error_name]
        except KeyError:
            raise ImproperlyConfigured(
                "Could not find remote exception in registry. "
                "Are you sure client and server ixmp versions are compatible?"
            )
        return exc.from_dict(json)

    def _request(self, method: str, path: str, *args, **kwargs) -> Mapping | Sequence:
        res = self.client.request(method, path, *args, **kwargs)

        if res.status_code >= 400:
            raise self.get_remote_exception(res, res.status_code)
        else:
            try:
                return res.json()
            except (ValueError, JSONDecodeError):
                return res.text

    def _get_by_id(self, id: int, *args, **kwargs) -> Mapping[str, Any]:
        # we can assume this type on create endpoints
        return self._request("GET", self.prefix + str(id) + "/", **kwargs)  # type: ignore

    def _enumerate(self, *args, table: bool = False, **kwargs) -> Mapping | Sequence:
        json = None
        params = {}
        params["table"] = table
        join_parameters = kwargs.pop("join_parameters", None)
        join_runs = kwargs.pop("join_runs", None)

        if join_parameters is not None:
            params["join_parameters"] = join_parameters
        if join_runs is not None:
            params["join_runs"] = join_runs

        if self.enumeration_method == "GET":
            params.update(kwargs)
        else:
            json = kwargs
        # we can assume this type on list endpoints
        return self._request(
            self.enumeration_method,
            self.prefix,
            params=self.sanitize_params(params),
            json=json,
        )

    def _list(self, **kwargs) -> Sequence[Mapping[str, Any]]:
        # we can assume this type on list endpoints
        return self._enumerate(table=False, **kwargs)  # type: ignore

    def _tabulate(self, **kwargs) -> pd.DataFrame:
        # we can assume this type on table endpoints
        jdf: Mapping[str, Any] = self._enumerate(table=True, **kwargs)  # type: ignore
        return DataFrame(**jdf).to_pandas()

    def _create(self, *args, **kwargs) -> Mapping[str, Any]:
        # we can assume this type on create endpoints
        return self._request("POST", *args, **kwargs)  # type: ignore

    def _delete(self, id: int):
        self._request("DELETE", self.prefix + str(id))


class Retriever(BaseRepository[ModelType]):
    def get(self, *args, **kwargs) -> ModelType:
        list_ = self._list(*args, **kwargs)

        try:
            [obj] = list_
        except ValueError:
            raise self.model_class.NotFound(
                f"Expected exactly one result, got {len(list_)} instead."
            )
        return self.model_class(**obj)


class Creator(BaseRepository[ModelType]):
    def create(self, **kwargs) -> ModelType:
        res = self._create(
            self.prefix,
            json=kwargs,
        )
        return self.model_class(**res)


class Deleter(BaseRepository[ModelType]):
    def delete(self, id: int):
        self._delete(id)


class Lister(BaseRepository[ModelType]):
    def list(self, *args, **kwargs) -> Iterable[ModelType]:
        return [
            self.model_class(**obj)
            for obj in self._list(
                **kwargs,
            )
        ]


class Tabulator(BaseRepository[ModelType]):
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return self._tabulate(*args, **kwargs)


class Enumerator(Lister[ModelType], Tabulator[ModelType]):
    def enumerate(
        self, *args, table: bool = False, **kwargs
    ) -> Iterable[ModelType] | pd.DataFrame:
        return self._enumerate(*args, table=table, **kwargs)


class BulkUpserter(BaseRepository[ModelType]):
    def bulk_upsert(self, df: pd.DataFrame, **kwargs) -> None:
        sdf = DataFrame.from_pandas(df)
        self._request(
            "POST",
            self.prefix + "bulk/",
            params=kwargs,
            data=sdf.json(),
        )


class BulkDeleter(BaseRepository[ModelType]):
    def bulk_delete(self, df: pd.DataFrame, **kwargs) -> None:
        sdf = DataFrame.from_pandas(df)
        self._request(
            "PATCH",
            self.prefix + "bulk/",
            params=kwargs,
            data=sdf.json(),
        )



from typing import Iterable, ClassVar
from datetime import datetime

import pandas as pd

from ixmp4.data import abstract
from .. import base


class DataPoint(base.BaseModel):
    NotFound: ClassVar = abstract.DataPoint.NotFound
    NotUnique: ClassVar = abstract.DataPoint.NotUnique
    DeletionPrevented: ClassVar = abstract.DataPoint.DeletionPrevented

    id: int
    time_series__id: int
    value: float
    type: str

    step_category: str | None
    step_year: int | None
    step_datetime: datetime | None


class DataPointRepository(
    base.Enumerator[DataPoint],
    base.BulkUpserter[DataPoint],
    base.BulkDeleter[DataPoint],
    abstract.DataPointRepository,
):
    model_class = DataPoint
    prefix = "iamc/datapoints/"
    enumeration_method = "PATCH"

    def list(self, *args, **kwargs) -> Iterable[DataPoint]:
        return super().list(*args, **kwargs)

    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    def enumerate(self, *args, **kwargs) -> Iterable[DataPoint] | pd.DataFrame:
        return super().enumerate(*args, **kwargs)

    def bulk_upsert(self, df: pd.DataFrame) -> None:
        super().bulk_upsert(df)

    def bulk_delete(self, df: pd.DataFrame) -> None:
        super().bulk_delete(df)



from datetime import datetime

from sqlalchemy.orm import (
    Mapped as Mapped,
)


Integer = Mapped[int]
Float = Mapped[float]
String = Mapped[str]
Boolean = Mapped[bool]
DateTime = Mapped[datetime]



# flake8: noqa
from .model import Model as Model
from .scenario import Scenario as Scenario
from .run import Run as Run
from .region import Region as Region
from .unit import Unit as Unit
from .platform import Platform as Platform
from .iamc.variable import Variable as Variable



from typing import Any
from fastapi import APIRouter, Depends, Query, Response, Request

from ixmp4.data.backend import Backend
from ixmp4.data import api
from ..base import BaseModel
from .. import deps


router: APIRouter = APIRouter(
    prefix="/timeseries",
    tags=["iamc", "timeseries"],
)


class TimeSeriesInput(BaseModel):
    run__id: int
    parameters: dict[str, Any]


class EnumerationOutput(BaseModel):
    __root__: list[api.TimeSeries] | api.DataFrame


@router.get("/", response_model=EnumerationOutput)
def enumerate(
    request: Request,
    join_parameters: bool | None = Query(False),
    run_ids: list[int] = Query(None),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    q = request.query_params
    parameters = {
        k: q[k] for k in q if k not in ["table", "run_ids", "join_parameters"]
    }
    kwargs = dict(
        run_ids=run_ids, join_parameters=join_parameters, parameters=parameters
    )
    return backend.iamc.timeseries.enumerate(table=bool(table), **kwargs)


@router.post("/", response_model=api.Run)
def create(
    timeseries: TimeSeriesInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.timeseries.create(**timeseries.dict())


@router.get("/{id}/", response_model=api.TimeSeries)
def get_by_id(
    id: int,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.timeseries.get_by_id(id)


@router.post("/bulk/")
def bulk_upsert(
    df: api.DataFrame,
    create_related: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    backend.iamc.timeseries.bulk_upsert(df.to_pandas(), create_related=create_related)
    return Response(status_code=201)



from typing import Iterable

import pandas as pd

from sqlalchemy.exc import NoResultFound

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.core.exceptions import OperationNotSupported

from .. import base

from .model import Unit
from .docs import UnitDocsRepository


class UnitRepository(
    base.Creator[Unit],
    base.Deleter[Unit],
    base.Retriever[Unit],
    base.Enumerator[Unit],
    abstract.UnitRepository,
):
    model_class = Unit

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        from .filter import UnitFilter

        self.filter_class = UnitFilter
        self.docs = UnitDocsRepository(*args, **kwargs)

    def add(self, name: str) -> Unit:
        if self.dialect.name == "oracle" and name.strip() == "":
            raise OperationNotSupported(
                "On ORACLE databases an empty string is not allowed as a unit's name."
            )
        unit = Unit(name=name, **self.get_creation_info())
        self.session.add(unit)
        return unit

    @guard("manage")
    def create(self, *args, **kwargs) -> Unit:
        return super().create(*args, **kwargs)

    @guard("manage")
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @guard("view")
    def get(self, name: str) -> Unit:
        exc: db.sql.Select = db.select(Unit).where(Unit.name == name)
        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Unit.NotFound

    @guard("view")
    def list(self, *args, **kwargs) -> Iterable[Unit]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)



from typing import Iterable, ClassVar
from datetime import datetime

import pandas as pd

from ixmp4.data import abstract
from .docs import DocsRepository, Docs
from . import base


class Scenario(base.BaseModel):
    NotFound: ClassVar = abstract.Scenario.NotFound
    NotUnique: ClassVar = abstract.Scenario.NotUnique
    DeletionPrevented: ClassVar = abstract.Scenario.DeletionPrevented

    id: int
    name: str
    created_at: datetime | None
    created_by: str | None


class ScenarioDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/scenarios/"


class ScenarioRepository(
    base.Creator[Scenario],
    base.Retriever[Scenario],
    base.Enumerator[Scenario],
    abstract.ScenarioRepository,
):
    model_class = Scenario
    prefix = "scenarios/"
    enumeration_method = "PATCH"

    def __init__(self, client, *args, **kwargs) -> None:
        super().__init__(client, *args, **kwargs)
        self.docs = ScenarioDocsRepository(self.client)

    def create(
        self,
        name: str,
    ) -> Scenario:
        return super().create(name=name)

    def get(self, name: str) -> Scenario:
        return super().get(name=name)

    def list(self, *args, **kwargs) -> Iterable[Scenario]:
        return super().list(*args, **kwargs)

    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    def enumerate(self, *args, **kwargs) -> Iterable[Scenario] | pd.DataFrame:
        return super().enumerate(*args, **kwargs)



# flake8: noqa
import importlib.metadata

from ixmp4.core import (
    Platform as Platform,
    Region as Region,
    Unit as Unit,
    Run as Run,
    Variable as Variable,
    Model as Model,
    Scenario as Scenario,
)

from ixmp4.core.exceptions import (
    NotFound as NotFound,
    NotUnique as NotUnique,
    IxmpError as IxmpError,
    InconsistentIamcType as InconsistentIamcType,
)

from ixmp4.data.abstract import DataPoint as DataPoint

__version__ = importlib.metadata.version("ixmp4")



from fastapi import APIRouter, Depends, Path, Query, Body

from ixmp4.data.backend import Backend
from ixmp4.data import api
from ixmp4.data.db.run.filter import RunFilter

from .base import BaseModel
from . import deps


router: APIRouter = APIRouter(
    prefix="/runs",
    tags=["runs"],
)


class RunInput(BaseModel):
    model_name: str
    scenario_name: str


class EnumerationOutput(BaseModel):
    __root__: list[api.Run] | api.DataFrame


@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: RunFilter = Depends(),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.runs.enumerate(table=bool(table), _filter=filter)


@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: RunFilter = Body(RunFilter()),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.runs.enumerate(table=bool(table), _filter=filter)


@router.post("/", response_model=api.Run)
def create(
    run: RunInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.runs.create(**run.dict())


@router.post("/{id}/set-as-default-version/")
def set_as_default_version(
    id: int = Path(None),
    backend: Backend = Depends(deps.get_backend),
):
    backend.runs.set_as_default_version(id)


@router.post("/{id}/unset-as-default-version/")
def unset_as_default_version(
    id: int = Path(None),
    backend: Backend = Depends(deps.get_backend),
):
    backend.runs.unset_as_default_version(id)



from typing import Iterable, ClassVar
import pandas as pd

from sqlalchemy.exc import NoResultFound

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.auth.decorators import guard
from ixmp4.data.abstract import iamc as abstract

from . import base
from .variable import Variable
from ..unit import Unit


class Measurand(base.BaseModel):
    NotFound: ClassVar = abstract.Measurand.NotFound
    NotUnique: ClassVar = abstract.Measurand.NotUnique
    DeletionPrevented: ClassVar = abstract.Measurand.DeletionPrevented

    __table_args__ = (db.UniqueConstraint("variable__id", "unit__id"),)

    variable__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("iamc_variable.id"), nullable=False, index=True
    )
    variable: types.Mapped[Variable] = db.relationship(
        "Variable",
        backref="measurands",
        foreign_keys=[variable__id],
        lazy="select",
    )

    unit__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("unit.id"), nullable=False, index=True
    )
    unit: types.Mapped[Unit] = db.relationship(
        "Unit",
        backref="measurands",
        foreign_keys=[unit__id],
        lazy="select",
    )

    created_at = db.Column(db.DateTime)
    created_by = db.Column(db.String(255))


class MeasurandRepository(
    base.Creator[Measurand],
    base.Retriever[Measurand],
    base.Enumerator[Measurand],
    abstract.MeasurandRepository,
):
    model_class = Measurand

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @guard("view")
    def get(self, variable_name: str, unit__id: int) -> Measurand:
        exc: db.sql.Select = (
            db.select(Measurand)
            .join(Measurand.variable)
            .where(Measurand.unit__id == unit__id)
            .where(Variable.name == variable_name)
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Measurand.NotFound

    def add(self, variable_name: str, unit__id: int) -> Measurand:
        q_exc: db.sql.Select = db.select(Variable).where(Variable.name == variable_name)
        try:
            variable = self.session.execute(q_exc).scalar_one()
        except NoResultFound:
            variable = Variable(name=variable_name)

        measurand = Measurand(variable=variable, unit__id=unit__id)
        self.session.add(measurand)
        self.session.commit()
        return measurand

    @guard("edit")
    def create(self, *args, **kwargs) -> Measurand:
        return super().create(*args, **kwargs)

    def select(self, *args, **kwargs) -> db.sql.Select:
        return db.select(Measurand)

    @guard("view")
    def list(self, *args, **kwargs) -> Iterable[Measurand]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)



from fastapi import APIRouter, Depends, Query, Body, Path

from ixmp4.data.backend import Backend
from ixmp4.data import api
from ixmp4.data.db.unit.filter import UnitFilter

from .decorators import autodoc
from .base import BaseModel
from . import deps


router: APIRouter = APIRouter(
    prefix="/units",
    tags=["units"],
)


class UnitInput(BaseModel):
    name: str


class EnumerationOutput(BaseModel):
    __root__: api.DataFrame | list[api.Unit]


@autodoc
@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: UnitFilter = Depends(),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.units.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: UnitFilter = Body(None),
    table: bool = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.units.enumerate(_filter=filter, table=bool(table))


@router.post("/", response_model=api.Unit)
def create(
    unit: UnitInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.units.create(**unit.dict())


@router.delete("/{id}/")
def delete(
    id: int = Path(None),
    backend: Backend = Depends(deps.get_backend),
):
    backend.units.delete(id)



"""
This module holds a shared datastructure and interface for normalization
between the database and api data models and repositories.
"""
# flake8: noqa

from .base import (
    BaseModel,
    BaseMeta,
    Retriever,
    Creator,
    Deleter,
    Lister,
    Tabulator,
    Enumerator,
    BulkUpserter,
    BulkDeleter,
)
from .model import Model, ModelRepository
from .run import Run, RunRepository
from .meta import RunMetaEntry, RunMetaEntryRepository, StrictMetaValue, MetaValue
from .scenario import Scenario, ScenarioRepository
from .unit import Unit, UnitRepository
from .region import Region, RegionRepository
from .docs import Docs, DocsRepository

from .iamc import (
    Variable,
    VariableRepository,
    Measurand,
    MeasurandRepository,
    TimeSeries,
    TimeSeriesRepository,
    DataPoint,
    #    AnnualDataPoint,
    #    SubAnnualDataPoint,
    #    CategoricalDataPoint,
    DataPointRepository,
)



from typing import Iterable, Any

from sqlalchemy import select
from sqlalchemy.orm import Bundle

import numpy as np
import pandas as pd
import pandera as pa
from pandera.typing import DataFrame, Series

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.unit import Unit
from ixmp4.data.db.region import Region
from ixmp4.data.db.run import RunRepository, Run
from ixmp4.data.db.model import Model
from ixmp4.data.db.scenario import Scenario
from ixmp4.core.exceptions import InconsistentIamcType, ProgrammingError
from ixmp4.core.decorators import check_types

from ..timeseries import TimeSeriesRepository, TimeSeries
from ..variable import Variable
from ..measurand import Measurand
from .. import base

from .model import DataPoint
from .filter import DataPointFilter
from . import get_datapoint_model


class RemoveDataPointFrameSchema(pa.SchemaModel):
    type: Series[pa.String] | None = pa.Field(isin=[t for t in DataPoint.Type])
    step_year: Series[pa.Int] | None = pa.Field(coerce=True, nullable=True)
    step_datetime: Series[pa.DateTime] | None = pa.Field(coerce=True, nullable=True)
    step_category: Series[pa.String] | None = pa.Field(nullable=True)

    time_series__id: Series[pa.Int] = pa.Field(coerce=True)

    @pa.dataframe_check
    def validate_type(cls, df: pd.DataFrame) -> bool:
        types = df.type.unique()

        # mixed types are currently not supported
        if len(types) != 1:
            raise InconsistentIamcType

        if types[0] == DataPoint.Type.ANNUAL:
            cols = dict(step_year="valid", step_category="empty", step_datetime="empty")
        elif types[0] == DataPoint.Type.CATEGORICAL:
            cols = dict(step_year="valid", step_category="valid", step_datetime="empty")
        elif types[0] == DataPoint.Type.DATETIME:
            cols = dict(step_year="empty", step_category="empty", step_datetime="valid")
        else:
            raise ProgrammingError

        for col, content in cols.items():
            if infer_content(df, col) is not content:
                raise InconsistentIamcType
        return True


class AddDataPointFrameSchema(RemoveDataPointFrameSchema):
    value: Series[pa.Float] = pa.Field(coerce=True)


class UpdateDataPointFrameSchema(AddDataPointFrameSchema):
    id: Series[pa.Int] = pa.Field(coerce=True)


def infer_content(df: pd.DataFrame, col: str) -> str:
    if col not in df.columns or all(pd.isna(df[col])):
        return "empty"
    elif not any(pd.isna(df[col])):
        return "valid"
    else:
        raise InconsistentIamcType


class DataPointRepository(
    base.Enumerator[DataPoint],
    base.BulkUpserter[DataPoint],
    base.BulkDeleter[DataPoint],
    abstract.DataPointRepository,
):
    model_class = DataPoint
    timeseries: TimeSeriesRepository
    runs: RunRepository

    def __init__(self, *args, **kwargs) -> None:
        backend, *_ = args
        # A different underlying database table
        # needs to be used for oracle databases.
        self.model_class = get_datapoint_model(backend.session)

        self.timeseries = TimeSeriesRepository(*args, **kwargs)
        self.runs = RunRepository(*args, **kwargs)

        self.filter_class = DataPointFilter
        super().__init__(*args, **kwargs)

    def join_auth(self, exc: db.sql.Select) -> db.sql.Select:
        if not db.utils.is_joined(exc, TimeSeries):
            exc = exc.join(
                TimeSeries, onclause=self.model_class.time_series__id == TimeSeries.id
            )
        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, TimeSeries.run)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return exc

    def select_joined_parameters(self, join_runs=False):
        _bundle = []
        if join_runs:
            _bundle.extend(
                [
                    Bundle("Model", Model.name.label("model")),
                    Bundle("Scenario", Scenario.name.label("scenario")),
                    Bundle("Run", Run.version),
                ]
            )

        _bundle.extend(
            [
                Bundle(
                    "Region",
                    Region.name.label("region"),
                ),
                Bundle(
                    "Unit",
                    Unit.name.label("unit"),
                ),
                Bundle(
                    "Variable",
                    Variable.name.label("variable"),
                ),
                self.bundle,
            ]
        )

        _exc = (
            select(*_bundle)
            .join(
                TimeSeries, onclause=self.model_class.time_series__id == TimeSeries.id
            )
            .join(Region, onclause=TimeSeries.region__id == Region.id)
            .join(Measurand, onclause=TimeSeries.measurand__id == Measurand.id)
            .join(Unit, onclause=Measurand.unit__id == Unit.id)
            .join(Variable, onclause=Measurand.variable__id == Variable.id)
        )

        if join_runs:
            _exc = (
                _exc.join(Run, onclause=TimeSeries.run__id == Run.id)
                .join(Model, onclause=Model.id == Run.model__id)
                .join(Scenario, onclause=Scenario.id == Run.scenario__id)
            )
        return _exc

    def select(
        self,
        *,
        join_parameters: bool | None = False,
        join_runs: bool = False,
        _filter: DataPointFilter | None = None,
        **kwargs: Any
    ) -> db.sql.Select:
        exc = (
            self.select_joined_parameters(join_runs)
            if join_parameters
            else select(self.bundle)
        )

        return super().select(_exc=exc, _filter=_filter, **kwargs)

    @guard("view")
    def list(self, *args, **kwargs) -> Iterable[DataPoint]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(
        self, *args: Any, _raw: bool | None = False, **kwargs: Any
    ) -> pd.DataFrame:
        if _raw:
            return super().tabulate(*args, **kwargs)
        df = super().tabulate(*args, **kwargs)
        df["value"] = df["value"].astype(np.float64)
        df = df.sort_values(
            by=["time_series__id", "step_year", "step_category", "step_datetime"]
        )
        return df.dropna(axis="columns", how="all")

    def check_df_access(self, df: pd.DataFrame):
        if self.backend.auth_context is not None:
            ts_ids = set(df["time_series__id"].unique().tolist())
            self.timeseries.check_access(ts_ids, access_type="edit")

    @check_types
    @guard("edit")
    def bulk_upsert(self, df: DataFrame[AddDataPointFrameSchema]) -> None:
        return super().bulk_upsert(df)

    @check_types
    @guard("edit")
    def bulk_insert(self, df: DataFrame[AddDataPointFrameSchema]) -> None:
        self.check_df_access(df)
        return super().bulk_insert(df)

    @check_types
    @guard("edit")
    def bulk_update(self, df: DataFrame[UpdateDataPointFrameSchema]) -> None:
        self.check_df_access(df)
        return super().bulk_update(df)

    @check_types
    @guard("edit")
    def bulk_delete(self, df: DataFrame[RemoveDataPointFrameSchema]) -> None:
        self.check_df_access(df)
        return super().bulk_delete(df)



from typing import Iterable, Optional, Any

from pydantic import BaseModel, Field, Extra, ValidationError

from ixmp4 import db
from ixmp4.core.exceptions import ProgrammingError, BadFilterArguments


def exact(c, v):
    return c == v


def in_(c, v):
    return c.in_(v)


def like(c, v):
    return c.like(v)


def ilike(c, v):
    return c.ilike(v)


def notlike(c, v):
    return c.notlike(v)


def notilike(c, v):
    return c.notilike(v)


def gt(c, v):
    return c > v


def lt(c, v):
    return c < v


def gte(c, v):
    return c >= v


def lte(c, v):
    return c <= v


class Integer(int):
    """An explicit proxy type for `int`."""

    pass


class Float(float):
    """An explicit proxy type for `float`."""

    pass


class Id(int):
    """A no-op type for a reduced set of `Integer` lookups."""

    pass


class String(str):
    """An explicit proxy type for `str`."""

    pass


Boolean = bool

argument_seperator = "__"
filter_func_prefix = "filter_"
lookup_map: dict[type, dict] = {
    Id: {
        "__root__": (int, exact),
        "in": (Iterable[int], in_),
    },
    Float: {
        "__root__": (int, exact),
        "in": (Iterable[int], in_),
        "gt": (int, gt),
        "lt": (int, lt),
        "gte": (int, gte),
        "lte": (int, lte),
    },
    Integer: {
        "__root__": (int, exact),
        "in": (Iterable[int], in_),
        "gt": (int, gt),
        "lt": (int, lt),
        "gte": (int, gte),
        "lte": (int, lte),
    },
    Boolean: {
        "__root__": (bool, exact),
    },
    String: {
        "__root__": (str, exact),
        "in": (Iterable[str], in_),
        "like": (Iterable[str], like),
        "ilike": (Iterable[str], ilike),
        "notlike": (Iterable[str], notlike),
        "notilike": (Iterable[str], notilike),
    },
}


def get_filter_func_name(n: str) -> str:
    return filter_func_prefix + n.strip()


PydanticMeta: type = type(BaseModel)


class FilterMeta(PydanticMeta):
    def __new__(cls, name, bases, namespace, **kwargs):
        field_types = namespace.get("__annotations__", {}).copy()
        for field_name, field_type in field_types.items():
            try:
                lookups = lookup_map[field_type]
            except KeyError:
                lookups = {}
                continue

            field = namespace.get(field_name, None)
            if field is not None:
                override_lookups = field.extra.get("lookups", None)
                if override_lookups:
                    lookups = {
                        k: v for k, v in lookups.items() if k in override_lookups
                    }
                elif override_lookups is None:
                    pass
                else:
                    lookups = {}
            cls.expand_lookups(
                field_name,
                lookups,
                namespace,
                None if field is None else field.alias,
            )

        return super().__new__(cls, name, bases, namespace, **kwargs)

    @classmethod
    def expand_lookups(
        cls, name, lookups, namespace, base_field_alias: str | None = None
    ):
        for lookup_alias, (type_, func) in lookups.items():
            if lookup_alias == "__root__":
                filter_name = name
            else:
                filter_name = name + argument_seperator + lookup_alias

            namespace["__annotations__"][filter_name] = Optional[type_]
            func_name = get_filter_func_name(filter_name)

            def filter_func(self, exc, f, v, func=func, session=None):
                return exc.where(func(f, v))

            namespace.setdefault(func_name, filter_func)

            field = namespace.get(filter_name, Field())
            field.extra.setdefault("sqla_column", name)
            if base_field_alias is not None and lookup_alias != "__root__":
                field.alias = base_field_alias + argument_seperator + lookup_alias
            namespace[filter_name] = field


class BaseFilter(BaseModel, metaclass=FilterMeta):
    class Config:
        extra = Extra.forbid
        sqla_model: type | None = None
        allow_population_by_field_name = True

    def __init__(self, **data: Any) -> None:
        try:
            super().__init__(**data)
        except ValidationError as e:
            raise BadFilterArguments(model=e.model, errors=e.errors)

    def join(self, exc: db.sql.Select, session=None) -> db.sql.Select:
        return exc

    def apply(self, exc: db.sql.Select, model, session) -> db.sql.Select:
        for name, field in self.__fields__.items():
            value = getattr(self, name, field.field_info.default)

            if isinstance(value, BaseFilter):
                submodel = getattr(value.Config, "sqla_model", None)
                model_getter = getattr(value.Config, "get_sqla_model", None)
                if submodel is None and callable(model_getter):
                    submodel = model_getter(session)

                exc = value.join(exc, session=session)
                exc = value.apply(exc, submodel, session)
            elif value is not None:
                func_name = get_filter_func_name(name)
                filter_func = getattr(self, func_name, None)
                if filter_func is None:
                    raise ProgrammingError

                sqla_column = field.field_info.extra.get("sqla_column")
                if sqla_column is None:
                    column = None
                else:
                    column = getattr(model, sqla_column, None)
                exc = filter_func(exc, column, value, session=session)
        return exc.distinct()



from fastapi import APIRouter, Depends, Query, Body

from ixmp4.data.backend import Backend
from ixmp4.data.db.scenario.filter import IamcScenarioFilter
from ixmp4.data import api

from ..base import BaseModel
from .. import deps

router: APIRouter = APIRouter(
    prefix="/scenarios",
    tags=["iamc", "scenarios"],
)


class EnumerationOutput(BaseModel):
    __root__: list[api.Scenario] | api.DataFrame


@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: IamcScenarioFilter = Depends(),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.scenarios.enumerate(table=bool(table), _filter=filter)


@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: IamcScenarioFilter = Body(IamcScenarioFilter()),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.scenarios.enumerate(table=bool(table), _filter=filter)



from fastapi import APIRouter, Depends, Path, Query

from ixmp4.data.backend import Backend
from ixmp4.data import api, abstract

from .base import BaseModel
from . import deps

router: APIRouter = APIRouter(
    prefix="/meta",
    tags=["meta"],
)


class RunMetaEntryInput(BaseModel):
    run__id: int
    key: str
    value: abstract.StrictMetaValue


class EnumerationOutput(BaseModel):
    __root__: api.DataFrame | list[api.RunMetaEntry]


@router.get("/", response_model=EnumerationOutput)
def enumerate(
    run_ids: list[int] | None = Query(None),
    keys: list[str] | None = Query(None),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.meta.enumerate(
        table=bool(table),
        run_ids=run_ids,
        keys=keys,
    )


@router.post("/", response_model=api.RunMetaEntry)
def create(
    runmeta: RunMetaEntryInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.meta.create(**runmeta.dict())


@router.delete("/{id}/")
def delete(
    id: int = Path(None),
    backend: Backend = Depends(deps.get_backend),
):
    backend.meta.delete(id)


@router.post("/bulk/")
def bulk_upsert(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.meta.bulk_upsert(df.to_pandas())


@router.patch("/bulk/")
def bulk_delete(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.meta.bulk_delete(df.to_pandas())



from fastapi import APIRouter, Depends, Query, Body

from ixmp4.data.backend import Backend
from ixmp4.data import api
from ixmp4.data.db.iamc.variable.filter import VariableFilter

from .. import deps
from ..decorators import autodoc
from ..base import BaseModel


router: APIRouter = APIRouter(
    prefix="/variables",
    tags=["iamc", "variables"],
)


class VariableInput(BaseModel):
    name: str


class EnumerationOutput(BaseModel):
    __root__: list[api.Variable] | api.DataFrame


@autodoc
@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: VariableFilter = Depends(),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.variables.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: VariableFilter = Body(VariableFilter()),
    table: bool = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.variables.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.post("/", response_model=api.Variable)
def create(
    variable: VariableInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.variables.create(**variable.dict())
